package CipUX::Import;

# author 2010 Christian Kuelker

# TODO: consider if this class also uses a configfile-loading role based on
# MooseX::ConfigFromFile, such as MooseX::SimpleConfig, MooseX::Getopt's
# new_with_options will load the configfile specified by the --configfile
# option (or the default you've given for the configfile attribute) for you.

# FIXME write more tests
# FIXME write POD
# FIXME add support for different line endings (slurp)
# FIXME find global approach to san regex
# FIXME find global approach to $s->get_import_share CONST
# FIXME add support for Debian
# FIXME add support for CipUX RPC --task (default), --rpc
# TODO add support for more switches
# TODO add support for more roles
# TODO add support for destroy accounts
# TODO add support for create class shares
# TODO add support for more share types
# TODO add support for different delimiter
# TODO add support for different formats
# TODO add support for --do-not-protect-me switch
#      (omitt line based ERRORS like #, empty line)
# TODO add support for --do-not-protect-me-at-all switch
#      (omitt all error checking)
use namespace::autoclean;
use Carp qw(confess croak);
use CipUX::Task;
use Data::Dumper;
use IO::File;
use IO::ScalarArray;
use Moose;
use MooseX::FollowPBP;
use MooseX::StrictConstructor;
use Perl6::Slurp;
use Readonly;
use Text::CSV;
use Text::CSV::Encoded;
use version; our $VERSION = qv('3.4.0.0');

with 'MooseX::Getopt';

# cipux_import --csv <FILE> --uid 1 --cipuxFirstname 2 --cipuxLastname 3 --cipuxRole 4
# --cipuxClassShare 5 --userPassword 6
# cipux_import --csv <FILE> --uid 1 --cipuxFirstname 2 --cipuxLastname 3 --cipuxRole 4
# --cipuxClassShare 5 --userPassword 6 --execute
has 'csv' => (
    is          => 'rw',
    isa         => 'Str',
    traits      => ['Getopt'],
    cmd_aliases => 'c',
    required    => 1
);
has 'execute' =>
    ( is => 'rw', isa => 'Bool', traits => ['Getopt'], cmd_aliases => 'e' );
has 'debug'           => ( is => 'rw', isa => 'Bool', traits => ['Getopt'] );
has 'verbose'         => ( is => 'rw', isa => 'Bool', traits => ['Getopt'] );
has 'uid'             => ( is => 'rw', isa => 'Num',  traits => ['Getopt'] );
has 'cipuxFirstname'  => ( is => 'rw', isa => 'Int',  traits => ['Getopt'] );
has 'cipuxLastname'   => ( is => 'rw', isa => 'Int',  traits => ['Getopt'] );
has 'cipuxRole'       => ( is => 'rw', isa => 'Int',  traits => ['Getopt'] );
has 'cipuxClassShare' => ( is => 'rw', isa => 'Int',  traits => ['Getopt'] );
has 'userPassword'    => ( is => 'rw', isa => 'Int',  traits => ['Getopt'] );

has 'column' => (
    is      => 'ro',
    isa     => 'ArrayRef',
    traits  => ['NoGetopt'],
    lazy    => 1,
    default => sub {
        [
            qw(uid cipuxFirstname cipuxLastname cipuxRole cipuxClassShare
                userPassword )
        ];
    }
);

has 'used_column_ar' => (
    is      => 'rw',
    isa     => 'ArrayRef',
    traits  => ['NoGetopt'],
    lazy    => 1,
    builder => '_used_column_builder'
);

# class lecture course reading studygroup team tutorial workshop seminar
has 'import_share' => (
    is      => 'ro',
    isa     => 'ArrayRef',
    traits  => ['NoGetopt'],
    default => sub {
        [
            qw(
                studygroup
                class
                course
                team
                workshop
                tutorial
                seminar
                lecture
                reading
                )
        ];
    }
);

has 'import_san_qr_hr' => (
    is      => 'ro',
    isa     => 'HashRef',
    traits  => ['NoGetopt'],
    default => sub {
        return {
            uid             => qr{^[a-z]+[a-z0-9]+$},
            cipuxFirstname  => qr{^[a-zA-Z0-9]+$},
            cipuxLastname   => qr{^[a-zA-Z0-9]+$},
            userPassword    => qr{^[a-zA-Z0-9-_!%]+$},
            cipuxRole       => [qw(teachers students admins no-internet)],
            cipuxClassShare => qr{^[a-z0-9]+$},
        };
    }
);

sub _used_column_builder {
    my ( $s, $i ) = @_;
    my %opt = ();

    $opt{uid}             = $s->get_uid;
    $opt{userPassword}    = $s->get_userPassword;
    $opt{cipuxLastname}   = $s->get_cipuxLastname;
    $opt{cipuxFirstname}  = $s->get_cipuxFirstname;
    $opt{cipuxRole}       = $s->get_cipuxRole;
    $opt{cipuxClassShare} = $s->get_cipuxClassShare;

    my %used_column = ();
    my $uc          = 0;
    foreach my $c ( @{ $s->get_column } ) {
        if ( exists $opt{$c} and defined $opt{$c} ) {
            $used_column{$c} = $opt{$c};
            $uc++;
        }
    }

    if ( $uc < 1 ) {
        croak "(EXCEPTION): give at least one more command line option --uid";
    }

    my @used_column
        = sort { $used_column{$a} <=> $used_column{$b} } keys %used_column;

    my $cl = join ' ', @used_column;
    print "Using columns: $cl\n" if defined $s->get_verbose;

    return \@used_column;
}

Readonly::Scalar my $EMPTY_STRING => q{};

my $import_hr = {};

my %opt = ();

my @column
    = qw(uid cipuxFirstname cipuxLastname cipuxRole cipuxClassShare userPassword);
my @all       = ();
my %all       = ();
my %all_role  = ();
my %all_share = ();

sub cipux_import {
    my ( $s, $i ) = @_;

    # untaint data captured by parens tainted
    no re 'taint';

    if ( $s->get_csv =~ m{^([[:alnum:]_\-/.]+)$}smx ) {
        $s->set_csv($1);    # data OK
    }
    elsif ( $s->get_csv eq $EMPTY_STRING ) {

    }
    else {
        my $msg = 'The name of the file is bad: ';
        $msg .= " [$s->get_csv]. \n";
        croak $msg;

    }

    # +-----------------------------------------------------------------
    # |  read CSV file in
    my @file = slurp $s->get_csv, { utf8 => 1 }, { irs => qr{\n+} };

    # +-----------------------------------------------------------------
    # |  first global check (line based)
    my $FH = new IO::ScalarArray \@file;

    my $n = 1;
    foreach my $l (@file) {
        my $lp = $l;
        chomp $lp;
        print "LINE $n [$lp]\n" if defined $s->get_debug;
        if ( $l =~ m{^$} ) {
            my $msg = "(ERROR): import file line $n is empty in $s->get_csv,";
            $msg .= " please remove it! (seems empty)\n";
            croak $msg;

        }
        elsif ( $l =~ m{^\#.*}smx ) {
            my $msg
                = "(ERROR): please remove import file line $n of $s->get_csv, ";
            $msg .= "because its not a CSV line! (garbage # at beginning)\n";
            croak $msg;
        }

 #  elsif ( $l =~ m{[^[:alnum:],'"]+}smx ) {
 #      my $msg = "(ERROR): line [$n] contains bad characters! [$s->get_csv]";
 #      croak $msg;
 #  }
        else {
        }
        $n++;
    }

    print "Input file has ", $n - 1, " entries.\n" if defined $s->get_verbose;

    # +-----------------------------------------------------------------
    # |  parse CSV file (column based)
    my $csv = Text::CSV::Encoded->new( { encoding => "utf8" } )
        or confess Text::CSV->error_diag();

    my $used_column_ar = $s->get_used_column_ar;
    foreach my $x ( @{$used_column_ar} ) {
    }
    $csv->column_names( @{$used_column_ar} );

    my $gl = 1;
    while ( my $hr = $csv->getline_hr($FH) ) {

        my @columnsx     = $csv->fields();
        my @a            = keys %{$hr};
        my $max_col_file = scalar @a;
        my $max_col_cl   = scalar @{$used_column_ar};

        if ( $max_col_file > scalar @{$used_column_ar} ) {
            croak "Input file has more columns then command line switches!\n";
        }
        if ( $max_col_file < scalar @{$used_column_ar} ) {
            croak "Input file has less columns then command line switches!\n";
        }

        if ( not defined $hr->{uid} ) {
            my $msg = "No uid (login) given in import file line $gl.";
            $msg .= " Or do you forgot to give --uid option?\n";
        }
        my $uid            = $hr->{uid};
        my $used_column_ar = $s->get_used_column_ar;
        foreach my $col ( @{$used_column_ar} ) {
            my $c = $s->san( $col, $hr->{$col}, $gl );
            $import_hr->{$uid}->{$col} = $c;
            if ( defined $s->get_debug and $col eq 'userPassword' ) {
                print " $col [$c]";
            }
            elsif ( defined $s->get_verbose and $col eq 'userPassword' ) {
                print " $col [use --debug to print]";
            }
            elsif ( $col ne 'userPassword'
                and ( defined $s->get_verbose or defined $s->get_debug ) )
            {
                print " $col [$c]";
            }
        }
        $import_hr->{$uid}->{line} = $gl;
        if ( not exists $import_hr->{$uid}->{count} ) {
            $import_hr->{$uid}->{count} = 1;
        }
        else {
            $import_hr->{$uid}->{count}++;
        }
        print " ($import_hr->{$uid}->{count}) \n"
            if defined $s->get_debug
                or defined $s->get_verbose;
        if ( $import_hr->{$uid}->{count} > 1 ) {
            croak
                "(EXCEPTION): double entry [$uid] in import file line $gl!\n";
        }

        $gl++;
        my $eof = $csv->eof();
        last if $eof;

    }

    # +-----------------------------------------------------------------
    # |  retrieve existing accounts
    my $task     = CipUX::Task->new();
    my $return_r = $task->task(
        {
            script  => __PACKAGE__,
            task    => 'cipux_task_list_user_accounts',
            mode    => 'cmd',
            object  => '',
            attr_hr => {},
        }
    );

    my @answer = keys %{ $return_r->{taskres_r} };
    foreach my $obj (@answer) {
        $all{$obj}++;
    }

    my %user_exist = map { $_ => 1 } @answer;

    # +-----------------------------------------------------------------
    # |  retrieve existing roles
    $return_r = $task->task(
        {
            script  => __PACKAGE__,
            task    => 'cipux_task_list_role_accounts',
            mode    => 'cmd',
            object  => '',
            attr_hr => {},
        }
    );
    @answer = keys %{ $return_r->{taskres_r} };
    foreach my $obj (@answer) {
        $all{$obj}++;
        $all_role{$obj}++;
    }
    my %role_exist = map { $_ => 1 } @answer;

    # +-----------------------------------------------------------------
    # |  retrieve existing shares
    foreach my $share ( @{ $s->get_import_share } ) {

        my $return_r = $task->task(
            {
                script  => __PACKAGE__,
                task    => 'cipux_task_list_' . $share . '_shares',
                mode    => 'cmd',
                object  => $share,
                attr_hr => {},
            }
        );

        my @answer = keys %{ $return_r->{taskres_r} };
        push @all, @answer;
        foreach my $obj (@answer) {
            $all{$obj}++;
            $all_share{$obj}++;
        }
    }

    # +-----------------------------------------------------------------
    # |  main
    $s->import_objects( { import_hr => $import_hr, mode => 'test' } );

    print "OK\n" if $s->get_verbose;

    if ( defined $s->get_execute ) {
        $s->import_objects(
            { import_hr => $import_hr, mode => 'explosive' } );
    }
    elsif ( defined $s->get_verbose or defined $s->get_debug ) {

        print '=' x 78 . "\n";
        print "If you would like to start this import specify --execute\n";
        print '=' x 78 . "\n";
    }

}

sub import_objects {

    my ( $s, $i ) = @_;

    my $import_hr
        = ( exists $i->{import_hr} )
        ? $i->{import_hr}
        : confess 'need param [import_hr] in call to sub';

    my $mode
        = ( exists $i->{mode} )
        ? $i->{mode}
        : confess 'need param [mode] in call to sub';

    # +-----------------------------------------------------------------
    # |  Recheck for missing field, double entries
    my %test = %all;
    foreach my $id ( sort keys %{$import_hr} ) {

        if ( $mode eq 'explosive' ) {
            print "=GO= " if defined $s->get_debug or defined $s->get_verbose;
        }
        else {
            print "TEST " if defined $s->get_debug or defined $s->get_verbose;
        }

        print "line [$import_hr->{$id}->{line}] ID [$id] "
            if defined $s->get_debug;
        if ( exists $all{$id} ) {
            my $msg = "(EXEPTION): object [$id] do already exists ";
            $msg .= "on the system. ";
            $msg
                .= "In the import file this is line $import_hr->{$id}->{line}!\n";
            croak $msg;
        }
        if ( exists $test{$id} ) {
            my $msg = "(EXEPTION): object [$id] do already exists ";
            $msg .= "in import file in line $import_hr->{$id}->{line}!\n";
            croak $msg;
        }

        my $attr_hr        = {};
        my $used_column_ar = $s->get_used_column_ar;
        foreach my $col ( @{$used_column_ar} ) {
            if ( not exists $import_hr->{$id}->{$col} ) {
                my $msg = "(EXEPTION): column $col claimed to be used, ";
                $msg
                    .= "but no data for it in line $import_hr->{$id}->{line}";
                croak $msg;
            }
            if ( not defined $import_hr->{$id}->{$col} ) {
                my $msg
                    = "(EXEPTION): column $col try to inject undef value, ";
                $msg .= "in line $import_hr->{$id}->{line}";
                croak $msg;
            }
            if ( $col eq 'userPassword' and defined $s->get_debug ) {
                print "$col [$import_hr->{$id}->{$col}] ";
            }
            elsif ( $col eq 'userPassword' and defined $s->get_verbose ) {
                print "$col [use --debug to see] ";
            }
            elsif ( $col ne 'userPassword' and defined $s->get_verbose ) {
                print "$col [$import_hr->{$id}->{$col}] ";
            }
            $attr_hr->{$col} = $import_hr->{$id}->{$col};

        }
        print "\n" if defined $s->get_verbose or $s->get_debug;

        # remember object
        $test{$id}++;

        if ( $mode eq 'explosive' ) {

            # damage;
            #  $attr_hr->{object} = $id;
            my $task     = CipUX::Task->new();
            my $return_r = $task->task(
                {
                    script  => __PACKAGE__,
                    task    => 'cipux_task_create_user_account',
                    mode    => 'shell',
                    object  => $id,
                    attr_hr => $attr_hr,
                }
            );
        }
        else {
            print "* TEST add account [$id]\n"
                if defined $s->get_debug
                    or defined $s->get_verbose;
        }

        if ( exists $import_hr->{$id}->{cipuxRole} ) {

            my @role = keys %all_role;
            if ( not grep {m/^$import_hr->{$id}->{cipuxRole}$/smx} @role ) {
                my $msg
                    = "class share [$import_hr->{$id}->{cipuxClassShare}]";
                $msg .= "do not exist! (line $import_hr->{$id}->{line})\n";
                croak $msg;
            }
            if ( $mode eq 'explosive' ) {
                my $task     = CipUX::Task->new();
                my $return_r = $task->task(
                    {
                        script  => __PACKAGE__,
                        task    => 'cipux_task_add_member_to_role_account',
                        mode    => 'shell',
                        object  => $import_hr->{$id}->{cipuxRole},
                        attr_hr => { value => $id },
                    }
                );
            }
            else {

                my $msg = "* TEST would add [$id] to role";
                $msg .= " [$import_hr->{$id}->{cipuxRole}]\n";
                print $msg
                    if defined $s->get_debug
                        or defined $s->get_verbose;

            }
        }

        if ( exists $import_hr->{$id}->{cipuxClassShare} ) {
            my @share = keys %all_share;
            if ( not grep {m/^$import_hr->{$id}->{cipuxClassShare}$/smx}
                @share )
            {
                my $msg
                    = "class share [$import_hr->{$id}->{cipuxClassShare}]";
                $msg .= "do not exist! (line $import_hr->{$id}->{line})\n";
                croak $msg;
            }

            if ( $mode eq 'explosive' ) {
                my $task     = CipUX::Task->new();
                my $return_r = $task->task(
                    {
                        script  => __PACKAGE__,
                        task    => 'cipux_task_add_member_to_class_share',
                        mode    => 'shell',
                        object  => $import_hr->{$id}->{cipuxClassShare},
                        attr_hr => { value => $id },
                    }
                );
            }
            else {
                my $msg = "* TEST would add [$id] to class";
                $msg .= " [$import_hr->{$id}->{cipuxClassShare}]\n";
                print $msg
                    if defined $s->get_debug
                        or defined $s->get_verbose;
            }

        }
    }

    return;
}

# from Web.pm
sub extract {

    my ( $s, $i ) = @_;

    my $answer_hr
        = ( exists $i->{answer_hr} )
        ? $i->{answer_hr}
        : confess 'need param [answer_hr] in call to sub';

    my @tpl_data = ();
    foreach my $obj ( keys %{ $answer_hr->{cmdres_r} } ) {

        # create an array of hash_refs
        my $data_hr = $answer_hr->{cmdres_r}->{$obj};
        my %data    = ();

        # key = uid, cipuxFirstname, cipuxLastname, ...
        foreach my $key ( keys %{$data_hr} ) {
            $data{$key} = join( ', ', @{ $data_hr->{$key} } );
            utf8::encode( $data{$key} );
        }
        push @tpl_data, \%data;
    }

    return \@tpl_data;
}

sub san {

    my $s = shift;                 # object
    my $c = shift;                 # col
    my $d = shift;                 # data
    my $l = shift || 'unknown';    # line

    my $san_hr = $s->get_import_san_qr_hr;

    if ( ref $san_hr->{$c} eq 'ARRAY' ) {

        my @res = grep {m/^$d$/smx} @{ $san_hr->{$c} };

        if ( scalar @res > 1 or scalar @res < 1 ) {
            print "\n";
            print Dumper( \@res );
            print Dumper( \@{ $san_hr->{$c} } );
            croak "data in line $l NOT OK: [$d] (not a valid $c)\n";

        }

        return $res[0];
    }
    else {
        if ( not $d =~ m/$san_hr->{$c}/smx ) {
            croak "Bad data [$d] found in line $l\n";
        }
        my $b = $d;
        $b =~ s/($san_hr->{$c})/$1/smx;
        return $b;

    }

}
1;
